#define CFG_ENTRIES 10

#include <cstdio>
#include <cstdlib>
#include <vector>
#include "defines.h"
#include "structs.h"
#include "Logger.h"
#include "Config.h"

Config::Config() {
}

Config::~Config() {
}

void Config::loadConfig() {
	Logger::add("Parsing server.cfg...\n");

	if(parseConfig() != 0) {
		Logger::add("server.cfg is either corrupt or missing, writing standard template.\n");
		writeConfig();
		parseConfig();
	}

	Logger::add("Configuration loaded successfully!\n\n");
}

int Config::getPlayerMax() {
	return playerMax;
}

int Config::getBackupInterval() {
	return backupInterval;
}

char *Config::getName() {
	return cfgName;
}

char *Config::getMotd() {
	return cfgMotd;
}

char *Config::getPort() {
	return cfgPort;
}

bool Config::getVerify() {
	return cfgVerify;
}

bool Config::getPublic() {
	return cfgPublic;
}

std::vector<User> Config::getUserList() {
	return userList;
}

std::vector<Banned> Config::getBannedList() {
	return bannedList;
}

void Config::writeConfig() {
	FILE *cfg;

	cfg = fopen("server.cfg", "w");
	fputs("# SERVERNAME\nname=Custom Server!\n", cfg);
	fputs("# WELCOME MESSAGE\nmotd=Welcome to my server!\n", cfg);
	fputs("# PORT\nport=25565\n", cfg);
	fputs("# MAX PLAYERS\nmax=16\n", cfg);
	fputs("# PUBLIC\npublic=true\n", cfg);
	fputs("# VERIFY NAMES\nverify-names=true\n", cfg);
	fputs("# BACKUP INTERVAL - MINUTES\nbackup=30\n", cfg);
	fputs("# PHYSICS - WATER\nwater=false\n", cfg);
	fputs("# PHYSICS - LAVA\nlava=false\n", cfg);
	fputs("# PHYSICS - SAND\nsand=false\n", cfg);
	fclose(cfg);
}

int Config::parseConfig() {
	int i;
	char *chr, *tmp, buf[64];
	FILE *cfg;

	cfg = fopen("server.cfg", "r");

	if(cfg == NULL)
		/* server.cfg isn't present. */
		return 1;

	for(i = 0; i < CFG_ENTRIES;) {
		fgets(buf, sizeof(buf), cfg);

		if(buf[0] != '#') {
			chr = strchr(buf, '=');

			if(chr == NULL) {
				/* Improper layout of a non-comment. */
				fclose(cfg);
				return 1;
			}

			chr++;

			tmp = strchr(buf, '\n');
			if(tmp) *tmp = 0;
			/* else We've reached the last entry. */

			if(strncmp(buf, "name", 4) == 0) {
				strncpy(cfgName, chr, sizeof(cfgName));
				i++;
			}

			else if(strncmp(buf, "motd", 4) == 0) {
				strncpy(cfgMotd, chr, sizeof(cfgMotd));
				i++;
			}

			else if(strncmp(buf, "port", 4) == 0) {
				strncpy(cfgPort, chr, sizeof(cfgPort));
				i++;
			}

			else if(strncmp(buf, "max", 3) == 0) {
				playerMax = atoi(chr);
				i++;
			}

			else if(strncmp(buf, "public", 6) == 0) {
				cfgPublic = strncmp(chr, "true", 4) ? false : true;
				i++;
			}

			else if(strncmp(buf, "verify-names", 12) == 0) {
				cfgVerify = strncmp(chr, "true", 4) ? false : true;
				i++;
			}

			else if(strncmp(buf, "backup", 6) == 0) {
				backupInterval = atoi(chr);
				backupInterval *= 60;			/* Seconds. */
				i++;
			}

			else if(strncmp(buf, "water", 5) == 0) {
				cfgPhysWater = strncmp(chr, "true", 4) ? false : true;
				i++;
			}

			else if(strncmp(buf, "lava", 4) == 0) {
				cfgPhysLava = strncmp(chr, "true", 4) ? false : true;
				i++;
			}

			else if(strncmp(buf, "sand", 4) == 0) {
				cfgPhysSand = strncmp(chr, "true", 4) ? false : true;
				i++;
			}
		}

		if(feof(cfg) && i < CFG_ENTRIES) {
			/* Configuration is corrupt, since i < CFG_ENTRIES and EOF has been reached. */
			fclose(cfg);
			return 1;
		}

	}

	fclose(cfg);
	parseUserList();
	parseBannedList();
	return 0;
}

void Config::parseUserList() {
	char userbuf[64], *chr;
	User user;
	FILE *userlist;

	userList.clear();

	userlist = fopen("users.txt", "r");

	if(userlist == NULL) {
		userlist = fopen("users.txt", "w");
		fputs("# Format follows:\n", userlist);
		fputs("# username:userlevel <- 255 is *max*\n", userlist);
		fputs("# Example: Notch:255\n", userlist);
		fclose(userlist);
		return;
	}

	while(feof(userlist) == 0) {
		fgets(userbuf, sizeof(userbuf), userlist);

		chr = strchr(userbuf, '\n');
		if(chr) *chr = 0;

		chr = strchr(userbuf, ':');
		if(chr == NULL) continue;

		*chr = 0;

		strcpy(user.name, userbuf);
		chr++;
		user.level = (unsigned char) atoi(chr);
		userList.push_back(user);
	}

	fclose(userlist);
}

void Config::parseBannedList() {
	char bannedbuf[64], *chr;
	struct Banned banned;
	FILE *banlist;

	bannedList.clear();

	banlist = fopen("banned.txt", "r");

	if(banlist == NULL) {
		banlist = fopen("banned.txt", "w");
		fputs("# Format follows:\n", banlist);
		fputs("# username:ip\n", banlist);
		fputs("# Remove IP, or set to zero for nameban only.\n", banlist);
		fclose(banlist);
		return;
	}

	while(feof(banlist) == 0) {
		fgets(bannedbuf, sizeof(bannedbuf), banlist);

		chr = strchr(bannedbuf, '\n');
		if(chr) *chr = 0;

		chr = strchr(bannedbuf, ':');
		if(chr == NULL) continue;

		*chr = 0;

		strcpy(banned.name, bannedbuf);
		chr++;
		strcpy(banned.ip, chr);
		bannedList.push_back(banned);
	}

	fclose(banlist);
}
