#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <climits>
#include <cstdarg>
#include <ctime>

#include <stack>
#include <queue>
#include <vector>

#include "defines.h"
#include "Config.h"
#include "Logger.h"
#include "Player.h"
#include "Level.h"
#include "md5.h"
#include "Various.h"
#include "structs.h"
#include "srvMessage.h"
#include "Server.h"

Server::Server() {
	unsigned char c;
	Config cfg;
	struct addrinfo hints;
	FILE *guestarea;

	playerNum = 0;
	timer = 0;
	time_beat = 0;
	time_savemap = 0;
	time_backupmap = 0;
	salt = 0;
	isOperated = false;
	currentlySending = 0;
	mapSend = NULL;

	threadsNeeded = 0;
	threadsRunning = 0;

	newmutex(mutexSendMap);
	newmutex(mutexPlayerList);
	newmutex(mutexThreadData);
	newmutex(mutexPhysicsQueue);

	memset(&timeout, 0, sizeof(struct timeval));
	for(c = UCHAR_MAX; c > 0; idList.push(--c));

#ifdef WIN32
	CreateDirectory("backups", NULL);
#else
	mkdir("backups", NULL);
#endif

	Logger::start();
	Logger::add("Initializing server!\n\n");

	/* Retrieve configuration. */
	cfg.loadConfig();

	playerMax = cfg.getPlayerMax();
	if(playerMax > 255)
		playerMax = 255;

	backupInterval = cfg.getBackupInterval();

	strcpy(srvName, cfg.getName());
	strcpy(srvMotd, cfg.getMotd());
	strcpy(srvPort, cfg.getPort());

	srvPublic = cfg.getPublic();
	srvVerify = cfg.getVerify();
	userList = cfg.getUserList();
	bannedList = cfg.getBannedList();
	/* ! Config */

	/* Guestarea */
	guestarea = fopen("guest.dat", "rb");

	if(guestarea == NULL) {
		memset(guestFrom, 0, sizeof(unsigned short) * 3);
		memset(guestTo, 0, sizeof(unsigned short) * 3);
		srvGuest = false;
	}

	else {
		fread(guestFrom, 2, 3, guestarea);
		fread(guestTo, 2, 3, guestarea);
		srvGuest = true;
		fclose(guestarea);
	}
	/* ! Guestarea */

	if(srvVerify == true) {
		srand((unsigned int) time(NULL));
		salt = rand();
	}

	lvl_restore = NULL;
	lvl = new Level();
	if(lvl->loadMap("server_level.dat", false) == false)
		exit(0);

	/* Initialize netcode... */
#ifdef WIN32
	WSAStartup(MAKEWORD(2, 2), &wsaData);
#endif

	heartbeat = (struct addrinfo*) malloc(sizeof(struct addrinfo));
	listener = (struct addrinfo*) malloc(sizeof(struct addrinfo));
	memset(&hints, 0, sizeof(struct addrinfo));

	/* HEARTBEAT ADDRINFO */
	hints.ai_family = PF_INET;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;
	getaddrinfo("www.minecraft.net", "80", &hints, &heartbeat);

	/* LISTENER ADDRINFO */
	hints.ai_flags = AI_PASSIVE;
	getaddrinfo(NULL, srvPort, &hints, &listener);

	/* Set up and bind a listener, add to fd_set master. */
	sock_listener = socket(listener->ai_family, listener->ai_socktype, listener->ai_protocol);
	bind(sock_listener, listener->ai_addr, listener->ai_addrlen);
	listen(sock_listener, 1);
	FD_ZERO(&master);
	FD_SET(sock_listener, &master);
	nfds = sock_listener;
}

Server::~Server() {
	delete lvl;
	closesock(sock_listener);
	freeaddrinfo(heartbeat);
	freeaddrinfo(listener);

#ifdef WIN32
	WSACleanup();
#endif

}

bool Server::step() {
	std::vector<Player*>::iterator iter;
	unsigned int i;
	int j;
	char buffer[PACKLEN_LOGIN];
	Player *player;

	time(&timer);

	if(timer > time_beat) {
		beat();
		time_beat = timer + 45;
	}

	if(timer > time_savemap) {
		lvl->saveMap("server_level.dat");
		Logger::add("Map saved! Players: %d/%d\n", playerNum, playerMax);
		time_savemap = timer + 60;
	}
	
	if(timer > time_backupmap) {

		if(playerNum > 0) {
			std::vector<Player*>::iterator iter;
			char backupFileName[64];
			struct tm *ts;
			FILE *playerListBackup;

			time(&timer);
			ts = localtime(&timer);

			sprintf(backupFileName, "backups\\backup-%d_%d_%d_%d", ts->tm_mon, ts->tm_mday, ts->tm_hour, ts->tm_min);
			lvl->saveMap(backupFileName);

			strcat(backupFileName, "-players.txt");
			playerListBackup = fopen(backupFileName, "w");
			
			lock(mutexPlayerList);
			for(iter = playerList.begin(); iter != playerList.end(); ++iter)
				fprintf(playerListBackup, "%s\n", (*iter)->name);
			unlock(mutexPlayerList);

			fclose(playerListBackup);

			Logger::add("Map backup: Saved.\n");
		}

		else 
			Logger::add("Map backup: No players online, skipping.\n"); 
        
		time_backupmap = timer + (backupInterval > 0 ? backupInterval : 60*5);
	}       

	/* Recieve from player sockets */
	/* Accept new connections */
	readfds = master;
	select(nfds+1, &readfds, 0, 0, &timeout);

	lock(mutexPlayerList);

	iter = playerList.begin();
	while(iter != playerList.end()) {
		player = (*iter);
		
		if(player->released == true) {
			iter = playerList.erase(iter);
			idList.push(player->id_byte);
			free(player);
			playerNum = playerList.size();
			isOperated = isOperatorOn();
			continue;
		}

		if(FD_ISSET(player->sock, &readfds)) {
			j = recv(player->sock, buffer, 1, MSG_PEEK);

			if(j < 1) {
				if(player->hidden == true)
					sendChatMsgLevelSelective(false, player->userlevel, player, "&e%s left the game!", player->name);
				else
					sendChatMsg(player, true, "&e%s left the game!", player->name);

				releasePlayer(player);
			}

			else {
				Task *task;

				task = (Task*) malloc(sizeof(Task));
				task->player = player;

				switch(buffer[0]) {
					case OPCODE_LOGIN:
						j = PACKLEN_LOGIN;
						task->Proc = &Server::handleLogin;
						break;
					case OPCODE_MOVE:
						j = PACKLEN_MOVE; // local call to parsePacket?
						task->Proc = &Server::parsePacket;
						break;
					case OPCODE_CHAT:
						j = PACKLEN_CHAT; // handleChat
						task->Proc = &Server::parsePacket;
						break;
					case OPCODE_BLOCK_IN:
						j = PACKLEN_BLOCK_IN; // local call?
						task->Proc = &Server::parsePacket;
						break;
					default:
						free(task);
						continue;
				}

				j = recv(player->sock, task->in_buf, j, 0);
				task->in_buf[j] = 0;

				lock(mutexThreadData);
				taskQueue.push(task);
				threadsNeeded = 4 + (playerNum/5);
				while(threadsRunning < threadsNeeded) {
					LaunchThread(NULL, &Server::thrExecTask, this);
					threadsRunning++;
				}
				unlock(mutexThreadData);
			}
		}
		
		iter++;
	}

	if(FD_ISSET(sock_listener, &readfds)) {
		struct sockaddr_in client_ip;

		player = (Player*) malloc(sizeof(Player));
		memset(player, 0, sizeof(Player));

		i = sizeof(client_ip);

		player->sock = accept(sock_listener, (struct sockaddr*) &client_ip, (socklen_t*) &i);
		strcpy(player->ip, inet_ntoa(client_ip.sin_addr));
		
		if(player->sock < 1)
			free(player);

		else {
			if(player->sock > nfds)
				nfds = player->sock;

			if(playerNum >= playerMax)
				kickPlayer(player, "Server is full!", false);

			else {
				playerList.push_back(player);
				FD_SET(playerList.back()->sock, &master);

				playerList.back()->id_byte = idList.top();
				idList.pop();

				playerList.back()->posx = lvl->getSpawnX();
				playerList.back()->posy = lvl->getSpawnY();
				playerList.back()->posz = lvl->getSpawnZ();
				playerList.back()->rotx = lvl->getRotX();
				playerList.back()->roty = lvl->getRotY();
				playerList.back()->actionBlock_which = 3;
				playerNum = playerList.size();
			}
		}
	}

	unlock(mutexPlayerList);

	msleep(5);
	return true;
}

bool Server::execTask() {
	Task *task;
	void (Server::*Proc) (Task*);

	lock(mutexThreadData);

	while(taskQueue.size() == 0) {
		if(threadsRunning > threadsNeeded) {
			threadsRunning--;
			unlock(mutexThreadData);
			return false;
		}
		unlock(mutexThreadData);
		msleep(50);
		lock(mutexThreadData);
	}

	task = taskQueue.front();
	taskQueue.pop();
	unlock(mutexThreadData);

	Proc = task->Proc;
	(this->*Proc) (task);
	free(task);

	return true;
}

void Server::handleLogin(Task *task) {
	char *loginbuffer, *chr;
	Player *player, *tmpPlayer;
	User *user;

	loginbuffer = task->in_buf;
	player = task->player;

	if(loginbuffer[1] != PROTO_VERSION) {
		releasePlayer(player);
		return;
	}

	loginbuffer += 2;
	chr = strchr(loginbuffer, 0x20);

	if(chr) *chr = 0;

	else {
		releasePlayer(player);
		return;
	}

	strcpy(player->name, loginbuffer);
	player->namelen = strlen(player->name);	// To avoid calling strlen(player->name) too many times.

	if(isBanned(player->name, player->ip) == true) {
		kickPlayer(player, "You are banned!", false);
		Logger::add("%s (%s) tried to connect, but is banned.\n", player->name, player->ip);
		return;
	}

	if(srvVerify == true) {
		loginbuffer += 64; // Bit hackish..
		loginbuffer[32] = 0;

		if(verifyName(player->name, loginbuffer) == false) {
			kickPlayer(player, "Illegal name.", false);
			Logger::add("%s (%s) tried to connect, but could not verify his name.\n", player->name, player->ip);
			return;
		}
	}

	tmpPlayer = getPlayerByName(player->name);

	if(tmpPlayer != NULL && tmpPlayer != player) {
		kickPlayer(tmpPlayer, "You logged in from another computer!", false); // Kick old one (could be timeout)
		Logger::add("%s (%s) logged in from another computer.\n", player->name, player->ip);
		return;
	}

	Logger::add("%s connected from IP: %s\n", player->name, player->ip);

	user = getUserByName(player->name);

	if(user != NULL) {
		player->userlevel = user->level;
		player->isoperator = (player->userlevel > 100);
		isOperated = isOperatorOn();
	}

	sendServerInfo(player);
	sendMap(player);
	sendPlayerList(player);

	/* WELCOME MESSAGES HERE */
	sendChatMsg(player, false, "&fType &c/help &ffor a list of commands!");
}

void Server::drawBlocks(unsigned short *drawFrom, unsigned short *drawTo, unsigned char block, bool overwriteDense, Player *player) {
	unsigned short tmp, tempFrom[3], tempTo[3], draw[3];
	unsigned char changedblock;
	int count = 0, packlen = 0;
	char drawbuf[256];

	if(drawFrom[0] < drawTo[0]) {
		tempFrom[0] = drawFrom[0];
		tempTo[0] = drawTo[0];
	}
		
	else {
		tempFrom[0] = drawTo[0];
		tempTo[0] = drawFrom[0];
	}

	if(drawFrom[1] < drawTo[1]) {
		tempFrom[1] = drawFrom[1];
		tempTo[1] = drawTo[1];
	}
		
	else { 
		tempFrom[1] = drawTo[1];
		tempTo[1] = drawFrom[1];
	}

	if(drawFrom[2] < drawTo[2]) {
		tempFrom[2] = drawFrom[2];
		tempTo[2] = drawTo[2];
	}
		
	else {
		tempFrom[2] = drawTo[2];
		tempTo[2] = drawFrom[2];
	}

	for(draw[0] = tempFrom[0]; draw[0] <= tempTo[0]; draw[0]++) {
		for(draw[1] = tempFrom[1]; draw[1] <= tempTo[1]; draw[1]++) {
			for(draw[2] = tempFrom[2]; draw[2] <= tempTo[2]; draw[2]++) {

				changedblock = lvl->getBlock(draw[0], draw[1], draw[2]);

				if(changedblock == block || ((changedblock == W_BLOCK_ROCK || changedblock > 0x31) && overwriteDense == false) ||
					(changedblock > 0x31 && player->userlevel < 201))  {
					count++;
					continue;
				}

				if(player->userlevel < 255 && count > ((player->userlevel-50)*3)) {
					broadcast(drawbuf, packlen, NULL);
					return;
				}

				drawbuf[packlen] = OPCODE_BLOCK_OUT;
				packlen++;

				tmp = htons(draw[0]);
				memcpy(drawbuf+packlen, &tmp, sizeof(short));
				packlen += sizeof(short);

				tmp = htons(draw[1]);
				memcpy(drawbuf+packlen, &tmp, sizeof(short));
				packlen += sizeof(short);

				tmp = htons(draw[2]);
				memcpy(drawbuf+packlen, &tmp, sizeof(short));
				packlen += sizeof(short);

				drawbuf[packlen] = block > 0x31 ? block - 0x31 : block;
				packlen++;

				lvl->setBlock(draw[0], draw[1], draw[2], block);

				if((sizeof(drawbuf) - packlen) >= PACKLEN_BLOCK_OUT)
					continue;

				broadcast(drawbuf, packlen, NULL);
				packlen = 0;

				msleep(sizeof(drawbuf) / 6);
			}
		}
	}

	if(packlen > 0)
		broadcast(drawbuf, packlen, NULL);
}

bool Server::changeBlockState(bool toDense, unsigned short *drawFrom, unsigned short *drawTo) {
	unsigned short tempFrom[3], tempTo[3], draw[3];
	unsigned char changedblock;

	if(drawFrom[0] < drawTo[0]) {
		tempFrom[0] = drawFrom[0];
		tempTo[0] = drawTo[0];
	}
		
	else {
		tempFrom[0] = drawTo[0];
		tempTo[0] = drawFrom[0];
	}

	if(drawFrom[1] < drawTo[1]) {
		tempFrom[1] = drawFrom[1];
		tempTo[1] = drawTo[1];
	}
		
	else { 
		tempFrom[1] = drawTo[1];
		tempTo[1] = drawFrom[1];
	}

	if(drawFrom[2] < drawTo[2]) {
		tempFrom[2] = drawFrom[2];
		tempTo[2] = drawTo[2];
	}
		
	else {
		tempFrom[2] = drawTo[2];
		tempTo[2] = drawFrom[2];
	}

	
	for(draw[0] = tempFrom[0]; draw[0] <= tempTo[0]; draw[0]++) {
		for(draw[1] = tempFrom[1]; draw[1] <= tempTo[1]; draw[1]++) {
			for(draw[2] = tempFrom[2]; draw[2] <= tempTo[2]; draw[2]++) {

				changedblock = lvl->getBlock(draw[0], draw[1], draw[2]);

				if(toDense == true) {
					if(changedblock <= 0x31 && changedblock > 0)
						changedblock += 0x31;
					else
						continue;
				}
				
				else {
					if(changedblock > 0x31)
						changedblock -= 0x31;
					else
						continue;
				}

				lvl->setBlock(draw[0], draw[1], draw[2], changedblock);
			}
		}
	}

	return true;
}

bool Server::pasteBlocks(unsigned short *drawFrom, unsigned short *drawTo, unsigned short *pasteLoc, bool samemap) {
	unsigned short tempFrom[3], tempTo[3], draw[3], tempLoc[3];
	char changedblock, block;
	Level *_lvl;

	_lvl = (samemap == true ? lvl : lvl_restore);

	if(drawFrom[0] < drawTo[0]) {
		tempFrom[0] = drawFrom[0];
		tempTo[0] = drawTo[0];
	}
		
	else {
		tempFrom[0] = drawTo[0];
		tempTo[0] = drawFrom[0];
	}

	if(drawFrom[1] < drawTo[1]) {
		tempFrom[1] = drawFrom[1];
		tempTo[1] = drawTo[1];
	}
		
	else { 
		tempFrom[1] = drawTo[1];
		tempTo[1] = drawFrom[1];
	}

	if(drawFrom[2] < drawTo[2]) {
		tempFrom[2] = drawFrom[2];
		tempTo[2] = drawTo[2];
	}
		
	else {
		tempFrom[2] = drawTo[2];
		tempTo[2] = drawFrom[2];
	}

	for(draw[0] = tempFrom[0]; draw[0] <= tempTo[0]; draw[0]++) {
		for(draw[1] = tempFrom[1]; draw[1] <= tempTo[1]; draw[1]++) {
			for(draw[2] = tempFrom[2]; draw[2] <= tempTo[2]; draw[2]++) {

				if(samemap == true) {
					tempLoc[0] = pasteLoc[0] + (draw[0] - tempFrom[0]);
					tempLoc[1] = pasteLoc[1] + (draw[1] - tempFrom[1]);
					tempLoc[2] = pasteLoc[2] + (draw[2] - tempFrom[2]);
				} 
				
				else {
					tempLoc[0] = draw[0];
					tempLoc[1] = draw[1];
					tempLoc[2] = draw[2];
				}


				changedblock = lvl->getBlock(draw[0], draw[1], draw[2]);
				if(_lvl->getBlock(tempLoc[0], tempLoc[1], tempLoc[2] ) == changedblock || (changedblock == 0x00 && samemap == true))
					continue;

				block = _lvl->getBlock(draw[0], draw[1], draw[2]);

				sendBlock(NULL, tempLoc[0], tempLoc[1], tempLoc[2], block, true);
				lvl->setBlock(tempLoc[0], tempLoc[1], tempLoc[2], block);
				msleep(2);
			}
		}
	}
		
	return true;
}

const char *Server::levelToClass(unsigned char userlevel) {

	if(userlevel == 255) {
		return "n &4Owner";
	}
	
	else if(userlevel > 0) {
		userlevel = (userlevel-1)/50;
		switch(userlevel) {
			case 0:
				return " &fUser";
			case 1:
				return " &aTrusted User";
			case 2:
				return "n &3OP";
			case 3:
				return " &9Trusted OP";
			case 4:
				return "n &5Admin";
			case 5:
				return "n &cElite";
		}
	}

	return " &7Guest";
}

void Server::setGuestArea() {
	FILE *guestarea;

	srvGuest = true;

	guestarea = fopen("guest.dat", "wb");
	fwrite(guestFrom, 2, 3, guestarea);
	fwrite(guestTo, 2, 3, guestarea);
	fclose(guestarea);
}

bool Server::isGuestBlock(unsigned short x, unsigned short y, unsigned short z) {
	if(srvGuest == false)
		return false;
	
	if(x < guestFrom[0] && x < guestTo[0])
		return false;
	if(y < guestFrom[1] && y < guestTo[1])
		return false;
	if(z < guestFrom[2] && z < guestTo[2])
		return false;

	if(x > guestFrom[0] && x > guestTo[0])
		return false;
	if(y > guestFrom[1] && y > guestTo[1])
		return false;
	if(z > guestFrom[2] && z > guestTo[2])
		return false;
	
	return true;
}

void Server::teleportPlayer(Player *toPlayer, Player *who) {
	char tpbuf[10];

	tpbuf[0] = 0x08;
	tpbuf[1] = -1;

	memcpy(tpbuf+2, &toPlayer->posx, sizeof(short));
	memcpy(tpbuf+4, &toPlayer->posy, sizeof(short));
	memcpy(tpbuf+6, &toPlayer->posz, sizeof(short));
	tpbuf[8] = toPlayer->rotx;
	tpbuf[9] = toPlayer->roty;
	send(who->sock, tpbuf, 10, 0);
	tpbuf[1] = who->id_byte;
	broadcast(tpbuf, 10, NULL);
}

void Server::fakeLog(Player *player,  bool out) {
	char logbuf[PACKLEN_SPAWN];

	if(out == true) {
		memset(logbuf, 0x20, PACKLEN_CHAT);

		logbuf[0] = 0x0D;
		logbuf[1] = 0x00;

		sendChatMsgLevelSelective(true, player->userlevel, player, "&e%s left the game!", player->name);
		sendChatMsgLevelSelective(false, player->userlevel, player, "&e%s is now hidden!", player->name);

		logbuf[0] = 0x0C;
		logbuf[1] = player->id_byte;
		broadcastLevelSelective(logbuf, 2, true, player->userlevel, player);
	}

	else {
		memset(logbuf, 0x20, PACKLEN_SPAWN);

		logbuf[0] = 0x07;
		logbuf[1] = player->id_byte;

		logbuf[2] = '&';
		logbuf[3] = getColorCode(player->userlevel);
		memcpy(logbuf+4, player->name, player->namelen);
		memcpy(logbuf+66, &player->posx, 2);
		memcpy(logbuf+68, &player->posy, 2);
		memcpy(logbuf+70, &player->posz, 2);
		logbuf[72] = player->rotx;
		logbuf[73] = player->roty;
		broadcastLevelSelective(logbuf, PACKLEN_SPAWN, true, player->userlevel, player);

		sendChatMsgLevelSelective(true, player->userlevel, player, "&e%s joined the game!", player->name);
		sendChatMsgLevelSelective(false, player->userlevel, player, "&e%s is no longer hidden!", player->name);
	}

}

bool Server::playerCommand(Player *player, char paramStr[3][64]) {
	char sprintfbuf[PACKLEN_CHAT];
	const char *msgptr;

	/* --- NOTE TO SELF --- */
	/* Many abbreviations (such as /p for /paint) */
	/* are still being called strcmp on. */
	/* This is unecessary, I could just compare them with a == */
	/* hmph, will probably do so in the future. */

	if(strcmp(paramStr[0], "me") == 0) {
		if(paramStr[1][0] == 0)
			return false;

		sendChatMsg(NULL, true, "* &b%s %s", player->name, paramStr[1]);
		return true;
	}

	else if(strcmp(paramStr[0], "set") == 0 || strcmp(paramStr[0], "setblock") == 0) {
		player->setblock = false;
		msgptr = srvMessage::BLOCK_NORMAL;

		if(paramStr[1] != NULL) {
			if(strcmp(paramStr[1], "g") == 0 || strcmp(paramStr[1], "grass") == 0) {
				player->block = W_BLOCK_GRASS;
				msgptr = srvMessage::BLOCK_GRASS;
				player->setblock = true;
			}

			else if(strcmp(paramStr[1], "l") == 0 || strcmp(paramStr[1], "lava") == 0) {
				player->block = W_BLOCK_SLAVA;
				msgptr = srvMessage::BLOCK_LAVA;
				player->setblock = true;
			}

			else if(strcmp(paramStr[1], "w") == 0 || strcmp(paramStr[1], "water") == 0) {
				player->block = W_BLOCK_SWATER;
				msgptr = srvMessage::BLOCK_WATER;
				player->setblock = true;
			}
		
			else if(strcmp(paramStr[1], "a") == 0 || strcmp(paramStr[1], "air") == 0) {
				player->block = W_BLOCK_AIR;
				msgptr = srvMessage::BLOCK_AIR;
				player->setblock = true;
			}
		}

		sendChatMsg(player, false, "%s%s", srvMessage::BLOCK_PLACING, msgptr);
		return true;
	}

	/* Thanks a lot to you, Dandistine, for helping us with such a beautiful solution. */
	/* Your help is much needed and appreciated.                      // vLK & Unison. */
	else if(strcmp(paramStr[0], "help") == 0) {
        unsigned int helpFrom = 0, helpTo = 8;

		if(paramStr[1][0] != 0) {
			if(strcmp(paramStr[1], "normal") == 0) {
				helpFrom = 9;
				helpTo = 13;
			}

			else if(strcmp(paramStr[1], "ranks") == 0) {
				helpFrom = 14;
				helpTo = 22;
			}

			else if(strcmp(paramStr[1], "build") == 0) {
				helpFrom = 23;
				helpTo = 36;
			}

			else if(strcmp(paramStr[1], "users") == 0) {
				helpFrom = 37;
				helpTo = 48;
			}
		}
		
		for(; helpFrom <= helpTo; helpFrom++)
			send(player->sock, srvMessage::HELP_MSG[helpFrom], PACKLEN_CHAT, 0);

		return true;        
	}

	else if(strcmp(paramStr[0], "pinfo") == 0) {
		Player *tmpPlayer;

		if(paramStr[1][0] == 0)
			return false;

		tmpPlayer = getPlayerByName(paramStr[1]);

		if(tmpPlayer != NULL)
	        sprintf(sprintfbuf, "&e%s is a%s &e(%d)", tmpPlayer->name, levelToClass(tmpPlayer->userlevel), tmpPlayer->userlevel);

		else {
			User *tmpUser;

			tmpUser = getUserByName(paramStr[1]);
			
			if(tmpUser != NULL)
				sprintf(sprintfbuf, "&e%s is a%s &e(%d)", tmpUser->name, levelToClass(tmpUser->level), tmpUser->level);

			else
				sprintf(sprintfbuf, "&eNo information found.");
		}

		sendChatMsg(player, false, sprintfbuf);
		return true;
    } 

	/* OP COMMANDS */
	if(player->userlevel < 51) {
		sendChatMsg(player, false, "&eUnknown command!");
		return false;
    }

	else if(strcmp(paramStr[0], "action") == 0 || strcmp(paramStr[0], "a") == 0) {
		player->action = ! player->action;

		if(player->action == true) {
			sendChatMsg(player, false, srvMessage::ACTION_INFO[0]);
			msgptr = srvMessage::ACTION_INFO[1];
		}

		else {
			sendBlock(player, player->actionBlock_from[0], player->actionBlock_from[1], player->actionBlock_from[2], 
				lvl->getBlock(player->actionBlock_from[0], player->actionBlock_from[1], player->actionBlock_from[2]), false);

			sendBlock(player, player->actionBlock_to[0], player->actionBlock_to[1], player->actionBlock_to[2], 
				lvl->getBlock(player->actionBlock_to[0], player->actionBlock_to[1], player->actionBlock_to[2]), false);

			player->actionBlock_which = 3;
			player->paste = false;
			msgptr = srvMessage::ACTION_INFO[2];
		}

		sendChatMsg(player, false, msgptr);
		return true;
	}


	else if(strcmp(paramStr[0], "draw") == 0 || strcmp(paramStr[0], "d") == 0) {
		if(player->actionBlock_which > 1) {
			sendChatMsg(player, false, srvMessage::ACTION_NOBLOCKS);
			return false;
		}

		sendChatMsg(player, false, srvMessage::ACTION_DRAW[0]);
		
		drawBlocks(player->actionBlock_from, player->actionBlock_to, player->lastblock, player->userlevel > 200, player);

		sendChatMsg(player, false, srvMessage::ACTION_DRAW[1]);

		sendBlock(player, player->actionBlock_from[0], player->actionBlock_from[1], player->actionBlock_from[2], W_BLOCK_LAVA, false);
		sendBlock(player, player->actionBlock_to[0], player->actionBlock_to[1], player->actionBlock_to[2], W_BLOCK_LAVA, false);
		return true;
	}
	
	if(strcmp(paramStr[0], "paint") == 0 || strcmp(paramStr[0], "p") == 0) {
		player->painting = ! player->painting;
		msgptr = srvMessage::ACTION_PAINT[player->painting == true ? 0 : 1];
		sendChatMsg(player, false, msgptr);
		return true;
	}
    
    else if(strcmp(paramStr[0], "user") == 0) {
		unsigned char level;
		Player *playerChanged;

		if(paramStr[1][0] == 0 || paramStr[2][0] == 0)
			return false;

		level = (unsigned char) atoi(paramStr[2]);

		if(player->userlevel < 255 && level > (player->userlevel - 50)) {
			sprintf(sprintfbuf, "&eYou cannot set user level higher than %d.", player->userlevel - 50);
			msgptr = (const char*) sprintfbuf;
		}

		else {
			playerChanged = getPlayerByName(paramStr[1]);
        
			if(setUser(paramStr[1], level, player) == false)
				msgptr = srvMessage::PLAYER_INFERIOR;

			else {
				sprintf(sprintfbuf, "&e%s %s%s &e(%d)", playerChanged ? playerChanged->name : paramStr[1], srvMessage::ACTION_OTHERCHANGE, levelToClass(level), level);
				msgptr = sprintfbuf;
			}
		}

		sendChatMsg(player, false, msgptr);
		return true;
	}

	else if(strcmp(paramStr[0], "tp") == 0) {
		Player *toPlayer;

		if(paramStr[0][1] == 0)
			return false;

		toPlayer = getPlayerByName(paramStr[1]);

		if(toPlayer == NULL || (toPlayer->hidden == true && toPlayer->userlevel > player->userlevel)) {
			sendChatMsg(player, false, srvMessage::PLAYER_NOT);
			return false;
		}

		teleportPlayer(toPlayer, player);
		return true;
	}

	if(player->userlevel < 101) {
		sendChatMsg(player, false, "&eUnknown command!");
		return false;
    }
	
    if(strcmp(paramStr[0], "kick") == 0) {
		Player *toKick;

		if(paramStr[1][0] == 0)
			return false;

		toKick = getPlayerByName(paramStr[1]);

		if(toKick == NULL) {
			sendChatMsg(player, false, srvMessage::PLAYER_NOT);
			return false;
		}

		if(toKick->userlevel > player->userlevel) {
			/* NOTIFY PLAYER - LACK OF PRIVILEDGE */
			return false;
		}

		kickPlayer(toKick, srvMessage::PLAYER_KICK, true);
		return true;
	}

	else if(strcmp(paramStr[0], "say") == 0) {
		if(paramStr[1][0] == 0)
			return false;

		sendChatMsg(NULL, true, "&e%s", paramStr[1]);
		return true;
	}

	else if(strcmp(paramStr[0], "mute") == 0) {
		Player *toMute;

		if(paramStr[1][0] == 0)
			return false;

		toMute = getPlayerByName(paramStr[1]);

		if(toMute == NULL) {
			sendChatMsg(player, false, srvMessage::PLAYER_NOT);
			return false;
		}

		if(player->userlevel <= toMute->userlevel) {
			/* NOTIFY PLAYER - LACK OF PRIVILEDGE */
			return false;
		}

		toMute->muted = ! toMute->muted;
		return true;
	}
	
	else if(strcmp(paramStr[0], "hide") == 0) {
		player->hidden = ! player->hidden;

		msgptr = (player->hidden == true ? srvMessage::ACTION_HIDE : srvMessage::ACTION_UNHIDE);

		fakeLog(player, player->hidden);
		sendChatMsg(player, false, msgptr);
		return true;
    }

	else if(strcmp(paramStr[0], "r") == 0 || strcmp(paramStr[0], "rock") == 0 || strcmp(paramStr[0], "solid") == 0) {
		if(player->block == W_BLOCK_ROCK) {
			player->setblock = false;
			player->block = W_BLOCK_STONE;
			msgptr = srvMessage::BLOCK_NORMAL;
		}

		else {
			player->setblock = true;
			player->block = W_BLOCK_ROCK;
			msgptr = srvMessage::BLOCK_ROCK;
		}

		sendChatMsg(player, false, "%s%s", srvMessage::BLOCK_PLACING, msgptr);
		return true;
    }

	else if(strcmp(paramStr[0], "bring") == 0) {
		Player *who;

		if(paramStr[1][0] == 0)
			return false;

		who = getPlayerByName(paramStr[1]);

		if(who == NULL) {
			sendChatMsg(player, false, srvMessage::PLAYER_NOT);
			return false;
		}

        if(who->userlevel > player->userlevel) {
			/* Notify it can't be done, no privilege. */
			return false;
		}

		teleportPlayer(player, who);
		return true;
	}
    
    if(player->userlevel < 151) {
		sendChatMsg(player, false, "&eUnknown command!");
		return false;
    }

	/* Nothing yet, maybe request backups? */
		
	if(player->userlevel < 201) { ///////////////////// ADMINS AND HIGHER
		sendChatMsg(player, false, "&eUnknown command!");
		return false;
    }

	if(strcmp(paramStr[0], "copy") == 0) {
		if(player->actionBlock_which > 1) {
			sendChatMsg(player, false, srvMessage::ACTION_NOBLOCKS);
			return false;
		}

		sendChatMsg(player, false, srvMessage::ACTION_COPY[2]);

		player->paste = true;
		return true;
	}

	else if(strcmp(paramStr[0], "harden") == 0) {
		if(player->actionBlock_which > 1) {
			sendChatMsg(player, false, srvMessage::ACTION_NOBLOCKS);
			return false;
		}

		sendChatMsg(player, false, srvMessage::ACTION_HARDEN[0]);
		changeBlockState(true, player->actionBlock_from, player->actionBlock_to);
		sendChatMsg(player, false, srvMessage::ACTION_HARDEN[1]);
		return true;
	}

	else if(strcmp(paramStr[0], "soften") == 0) {
		if(player->actionBlock_which > 1) {
			sendChatMsg(player, false, srvMessage::ACTION_NOBLOCKS);
			return false;
		}

		sendChatMsg(player, false, srvMessage::ACTION_SOFTEN[0]);
		changeBlockState(false, player->actionBlock_from, player->actionBlock_to);
		sendChatMsg(player, false, srvMessage::ACTION_SOFTEN[1]);
		return true;
	}

	else if(strcmp(paramStr[0], "dense") == 0) {
		player->densing = ! player->densing;
		msgptr = srvMessage::ACTION_DENSE[player->densing == true ? 0 : 1];
		sendChatMsg(player, false, msgptr);
		return true;
	}

	else if(strcmp(paramStr[0], "setspawn") == 0)
		return lvl->setSpawn(player->posx, player->posy, player->posz, player->rotx, player->roty);

	/* Will always ban IP. */

    else if(strcmp(paramStr[0], "ban") == 0) {
		Player *toBan;

		if(paramStr[1][0] == 0)
			return false;

		toBan = getPlayerByName(paramStr[1]);

		if(toBan != NULL) {
			if(toBan->userlevel > player->userlevel)
				return false;
		}

		banPlayer(paramStr[1], srvMessage::PLAYER_BAN, false);
		return true;
	}

	else if(strcmp(paramStr[0], "unban") == 0) {
		if(paramStr[1][0] == 0)
			return false;

		banPlayer(paramStr[1], NULL, true);
		return true;
	}

	if(player->userlevel < 255) {
		sendChatMsg(player, false, "&eUnknown command!");
		return false;
	}

	if(strcmp(paramStr[0], "guest") == 0) {
		if(player->actionBlock_which > 1)  {
			sendChatMsg(player, false, srvMessage::ACTION_NOBLOCKS);
			return false;
		}

		sendChatMsg(player, false, srvMessage::ACTION_GUEST);

		memcpy(&guestFrom, &player->actionBlock_from, 6);
		memcpy(&guestTo, &player->actionBlock_to, 6);

		setGuestArea();
		return true;
	}

	else if(strcmp(paramStr[0], "clearguest") == 0 || strcmp(paramStr[0], "cg") == 0) {
		if(srvGuest == false)
			return false;

		drawBlocks(guestFrom, guestTo, W_BLOCK_AIR, false, player);
		return true;
	}

	else if(strcmp(paramStr[0], "restore") == 0) {
		if(player->actionBlock_which > 1) {
			sendChatMsg(player, false, srvMessage::ACTION_NOBLOCKS);
			return false;
		}

		if(paramStr[1][0] == 0)
			return false;

		if(lvl_restore == NULL)
			lvl_restore = new Level;
			
		if(lvl_restore->getServerLevel() != NULL || lvl_restore->loadMap(paramStr[1], true) == false)
			msgptr = srvMessage::ACTION_RESTORE[2];

		else {
			sendChatMsg(player, false, srvMessage::ACTION_RESTORE[0]);
				
			pasteBlocks(player->actionBlock_from, player->actionBlock_to, player->actionBlock_from, false);

			msgptr = srvMessage::ACTION_RESTORE[1];

			sendBlock(player, player->actionBlock_from[0], player->actionBlock_from[1], player->actionBlock_from[2], W_BLOCK_LAVA, false);
			sendBlock(player, player->actionBlock_to[0], player->actionBlock_to[1], player->actionBlock_to[2], W_BLOCK_LAVA, false);
		}

		if(lvl_restore) {
			delete lvl_restore;
			lvl_restore = NULL;
		}

		sendChatMsg(player, false, msgptr);
		return true;
	}

	sendChatMsg(player, false, "&eUnknown command!");
	return false;
}

void Server::releasePlayer(Player *player) {
	std::vector<Player*>::iterator iter;
	Player *tmp;
	char buffer[128];

	lock(mutexPlayerList);

	for(iter = playerList.begin(); iter != playerList.end(); ++iter) {

		tmp = (*iter);

		if(tmp == player) {
			Logger::add("%s (%s) has disconnected.\n", tmp->name, tmp->ip);

			buffer[0] = 0x0C;
			buffer[1] = tmp->id_byte;

			if(tmp->hidden == true)
				broadcastLevelSelective(buffer, 2, false, player->userlevel, player);
			else
				broadcast(buffer, 2, NULL);

			FD_CLR(tmp->sock, &master);

			/* Cannot be done here at the moment. */
			tmp->released = true;
			closesock(tmp->sock);
			break;
		}
	}

	unlock(mutexPlayerList);
}

void Server::kickPlayer(Player *player, const char *kickMsg, bool announce) {
	char kickbuf[PACKLEN_KICK];

	memset(kickbuf, 0x20, PACKLEN_KICK);

	kickbuf[0] = 0x0E;
	memcpy(kickbuf+1, kickMsg, strlen(kickMsg));
	send(player->sock, kickbuf, PACKLEN_KICK, 0);

	if(announce)
		sendChatMsg(NULL, true, "&e%s got kicked from the server!", player->name);

	releasePlayer(player);
}

void Server::banPlayer(char *playerName, const char *banMsg, bool unban) {
	std::vector<Banned>::iterator iter;
	bool alreadyBanned;
	struct Banned banned;
	Player *toBan;
	FILE *banlist;

	toBan = getPlayerByName(playerName);

	if(toBan != NULL) {
		strcpy(banned.ip, toBan->ip);
		kickPlayer(toBan, banMsg, false);
		sendChatMsg(NULL, true, "&e%s got banned!", toBan->name);
	}

	else banned.ip[0] = 0;

	strcpy(banned.name, playerName);
	toLowerCase(banned.name);

	alreadyBanned = isBanned(banned.name, banned.ip);

	/* Do not ban twice. */
	if(unban == false && alreadyBanned == true)
		return;
	if(unban == true && alreadyBanned == false)
		return;

	if(unban == true) {
		for(iter = bannedList.begin(); iter != bannedList.end(); iter++) {
			if(strcmp(iter->name, banned.name) == 0) {
				bannedList.erase(iter);
				break;
			}
		}
	}

	else
		bannedList.push_back(banned);

	banlist = fopen("banned.txt", "w");
	
	for(iter = bannedList.begin(); iter != bannedList.end(); iter++) {
		if(iter->name[0] != 0 || iter->ip[0] != 0)
			fprintf(banlist, "%s:%s\n", iter->name, iter->ip);
	}

	fclose(banlist);
}

User *Server::getUserByName(char *playerName) {
	char playerCompare[32];
	std::vector<User>::iterator iter;

	strcpy(playerCompare, playerName);
	toLowerCase(playerCompare);
	
	for(iter = userList.begin(); iter != userList.end(); iter++) {
		if(strcmp(iter->name, playerCompare) == 0)
			return &(*iter);
	}

	return NULL;
}

void Server::refreshColor(Player *player) {
	char logbuf[PACKLEN_SPAWN];
		
	logbuf[0] = 0x0C;
	logbuf[1] = player->id_byte;
	broadcast(logbuf, 2, player);
        
	memset(logbuf, 0x20, PACKLEN_SPAWN);

	logbuf[0] = 0x07;
	logbuf[1] = player->id_byte;

	logbuf[2] = '&';
	logbuf[3] = getColorCode(player->userlevel);
	memcpy(logbuf+4, player->name, player->namelen);
	memcpy(logbuf+66, &player->posx, 2);
	memcpy(logbuf+68, &player->posy, 2);
	memcpy(logbuf+70, &player->posz, 2);
	logbuf[72] = player->rotx;
	logbuf[73] = player->roty;
	broadcast(logbuf, PACKLEN_SPAWN, player);
}

bool Server::setUser(char *playerName, unsigned char level, Player *changer) {
	std::vector<User>::iterator iter;
	char playerCompare[32];
	User *user;
	Player *player;
	FILE *userlist;

	strcpy(playerCompare, playerName);
	toLowerCase(playerCompare);

	user = getUserByName(playerCompare);

	if(user != NULL) {
		if(changer->userlevel != 255 && changer->userlevel <= user->level) {
			/* Notify changer this can not be done. */
			return false;
		}
		user->level = level;
	}

	else {
		User tmp;
		strcpy(tmp.name, playerCompare);
		tmp.level = level;
		userList.push_back(tmp);
	}

	player = getPlayerByName(playerCompare);

	if(player != NULL) {
		player->userlevel = level;
		player->isoperator = (player->userlevel > 100);
		sendChatMsg(player, false, "%s%s &e(%d)", srvMessage::ACTION_LEVELCHANGE, levelToClass(player->userlevel), player->userlevel);
		refreshColor(player);
	}

	userlist = fopen("users.txt", "w");

	for(iter = userList.begin(); iter != userList.end(); iter++) {
		if(iter->name[0] != 0)
			fprintf(userlist, "%s:%d\n", iter->name, iter->level);
	}

	fclose(userlist);
	return true;
}

bool Server::verifyName(char *name, char *hash) {
	unsigned char sig[16];
	char hashed[33], hashbuf[64];

	sprintf(hashbuf, "%d%s", salt, name);
	md5_buffer(hashbuf, strlen(hashbuf), sig);
	md5_sig_to_string(sig, hashed, sizeof(hashed));

	if(strcmp(hash, hashed) == 0)
		return true;

	return false;
}

void Server::beat() {
	int sock;
	char httpheader[300], hbeat[256];
	static bool firstBeat = true;

	sprintf(hbeat, "port=%s&users=%d&max=%d&name=%s&public=%s&version=7&salt=%d", srvPort, playerNum, playerMax, srvName, srvPublic ? "true" : "false", salt);
	sprintf(httpheader, "POST /heartbeat.jsp HTTP/1.1\r\nContent-Type:application/x-www-form-urlencoded\r\nContent-Length:%d\r\nContent-Language: en-US\r\nCache-Control: no-cache\r\nPragma:no-cache\r\nUser-Agent: minerCPP\r\nHost:www.minecraft.net\r\nAccept: text/html, image/gif, image/jpeg, *; q=.2,*/*; q=.2\r\nConnection: keep-alive\r\n\r\n", strlen(hbeat));

    sock = socket(heartbeat->ai_family, heartbeat->ai_socktype, heartbeat->ai_protocol);
    connect(sock, heartbeat->ai_addr, heartbeat->ai_addrlen);
    send(sock, httpheader, strlen(httpheader), 0);
    send(sock, hbeat, strlen(hbeat), 0);

	if(firstBeat == true) {
		char hbeatrecv[300];
		char *beatptr;
		const char *msgptr;
		FILE *exurl;

		recv(sock, hbeatrecv, sizeof(hbeatrecv), 0);
		beatptr = strstr(hbeatrecv, "http://www.minecraft.net");

		if(beatptr != NULL) {
			exurl = fopen("external_url.txt", "w");
			fwrite(beatptr, 1, 73, exurl);
			fclose(exurl);
			msgptr = "Ext. URL saved to file: external_url.txt\n";
			firstBeat = false;
		}

		else
			msgptr = "Couldn't retrieve ext. URL. Will try again on next heartbeat.\n";

		Logger::add(msgptr);
	}

	closesock(sock);
}

bool Server::isOperatorOn() {
	std::vector<Player*>::iterator iter;
	bool ret = false;

	lock(mutexPlayerList);

	for(iter = playerList.begin(); iter != playerList.end(); ++iter) {
		if((*iter)->isoperator == true) {
			ret = true;
			break;
		}
	}

	unlock(mutexPlayerList);
	return ret;
}

/* Exclude can be used for sending stuff to everyone but "this one player". */
void Server::broadcast(char *packet, int paklen, Player *exclude) {
	std::vector<Player*>::iterator iter;
	
	lock(mutexPlayerList);

	for(iter = playerList.begin(); iter != playerList.end(); ++iter) {
		if((*iter) == exclude)
			continue;

		send((*iter)->sock, packet, paklen, 0);
	}

	unlock(mutexPlayerList);
}

/* Send to lowers OR higher. */
void Server::broadcastLevelSelective(char *packet, int paklen, bool toLowers, unsigned char level, Player *exclude) {
	std::vector<Player*>::iterator iter;

	lock(mutexPlayerList);

	for(iter = playerList.begin(); iter != playerList.end(); ++iter) {
		if((*iter) == exclude || (! toLowers == (*iter)->userlevel < level))
			continue;

		send((*iter)->sock, packet, paklen, 0);
	}

	unlock(mutexPlayerList);
}

bool Server::opchat(Player *player, char *buffer) {
	if(buffer[0] == 0)
		return false;

	sendChatMsgLevelSelective(false, 101, NULL, "#&%c%s&f: %s", getColorCode(player->userlevel), player->name, buffer);
	return true;
}

bool Server::privsay(Player *player, char *buffer) {
	char *chr;
	Player *toPlayer;

	if(buffer[0] == 0x20) {
		buffer++;
		if(buffer[0] == 0) {
			/* Player has requested to flush his id_privsay... */
			player->id_privsay = player->id_byte;
			return false;
		}
	}

	chr = strchr(buffer, 0x20);
	if(chr) *chr = 0;
	else return false;

	toPlayer = getPlayerByName(buffer);

	if(toPlayer == NULL) {
		sendChatMsg(player, false, srvMessage::PLAYER_NOT);
		return false;
	}

	player->id_privsay = toPlayer->id_byte;	/* This should allow higher ranked, hidden players to initialize a char with lowers. */

	if(toPlayer == player || (toPlayer->hidden == true && player->userlevel < toPlayer->userlevel && player->id_byte != toPlayer->id_privsay)) {
		/* Do not pm yourself, faggot. */
		/* Nor somebody of a higher userlevel, that is currently hidden :D */
		sendChatMsg(player, false, srvMessage::PLAYER_NOT);
		return false;
	}

	chr++;

	if(chr[0] != 0) {
		sendChatMsg(player, false, "&6(&%c%s&6) >>&7 %s", getColorCode(toPlayer->userlevel), toPlayer->name, chr);
		sendChatMsg(toPlayer, false, "&6(&%c%s&6) :&7 %s", getColorCode(player->userlevel), player->name, chr);
	}

	return true;
}

char Server::getColorCode(unsigned char userlevel) {
	if(userlevel == 255)
		return '4';
	if(userlevel > 250)
		return 'c';
	else if(userlevel > 200)
		return '5';
	else if(userlevel > 150)
		return '9';
	else if(userlevel > 100)
		return '3';
	else if(userlevel > 50)
		return 'a';
	else if(userlevel > 0)
		return 'f';
	return '7';
}

void Server::stripChatPacket(char *packet) {
	int i;
	char *chr;

	/* PACKLEN_CHAT-1 ; [PACKLEN_CHAT] is the null termination */
	for(chr = packet+PACKLEN_CHAT-1; *chr == 0x20; chr--);
	*(++chr) = 0x00;

	/* strlen + 1 ; in order to pick up the null termination. */
	i = strlen(packet+2) + 1;
	memmove(packet, packet+2, i);
}

void Server::sendChatMsg(Player *player, bool broad, const char *format, ...) {
	int len;
	char chatbuf[PACKLEN_CHAT], vstr[128];
	va_list args;

	va_start(args, format);
	vsprintf(vstr, format, args);
	va_end(args);

	len = strlen(vstr);

	memset(chatbuf, 0x20, PACKLEN_CHAT);
	chatbuf[0] = OPCODE_CHAT;
	chatbuf[1] = 0x00;
	memcpy(chatbuf+2, vstr, len < PACKLEN_CHAT-2 ? len : PACKLEN_CHAT-2);

	if(broad == true)
		broadcast(chatbuf, PACKLEN_CHAT, player);
	else
		send(player->sock, chatbuf, PACKLEN_CHAT, 0);
}

void Server::sendChatMsgLevelSelective(bool toLowers, unsigned char level, Player *exclude, const char *format, ...) {
	int len;
	char chatbuf[PACKLEN_CHAT], vstr[128];
	va_list args;

	va_start(args, format);
	vsprintf(vstr, format, args);
	va_end(args);

	len = strlen(vstr);

	memset(chatbuf, 0x20, PACKLEN_CHAT);
	chatbuf[0] = OPCODE_CHAT;
	chatbuf[1] = 0x00;
	memcpy(chatbuf+2, vstr, len < PACKLEN_CHAT-2 ? len : PACKLEN_CHAT-2);

	broadcastLevelSelective(chatbuf, PACKLEN_CHAT, toLowers, level, exclude);
}

void Server::sendBlock(Player *player, unsigned short x, unsigned short y, unsigned short z, unsigned char block, bool broad) {
	unsigned short tmp;
	char buffer[PACKLEN_BLOCK_OUT];
	
	buffer[0] = OPCODE_BLOCK_OUT;

	tmp = htons(x);
	memcpy(buffer+1, &tmp, sizeof(short));
	
	tmp = htons(y);
	memcpy(buffer+3, &tmp, sizeof(short));
	
	tmp = htons(z);
	memcpy(buffer+5, &tmp, sizeof(short));
	
	if(block > 0x31)
		block -= 0x31;

	buffer[7] = block;

	if(broad == true)
		broadcast(buffer, PACKLEN_BLOCK_OUT, NULL);
	else
		send(player->sock, buffer, PACKLEN_BLOCK_OUT, 0);
}

/* Parses packets sent by the clients */
/* and performs all needed actions */
void Server::parsePacket(Task *task) {
	int i;
	char *buffer;
	Player *player;

	buffer = task->in_buf;
	player = task->player;
	
	if(buffer[0] == OPCODE_MOVE) {
		buffer[1] = player->id_byte;

		memcpy(&player->posx, buffer+2, 2);
		memcpy(&player->posy, buffer+4, 2);
		memcpy(&player->posz, buffer+6, 2);
		player->rotx = buffer[8];
		player->roty = buffer[9];

		if(player->hidden == true)
			broadcastLevelSelective(buffer, PACKLEN_MOVE, false, player->userlevel, player);
		else
			broadcast(buffer, PACKLEN_MOVE, player);
	}

	/* Blockchange */
	else if(buffer[0] == OPCODE_BLOCK_IN) {
		unsigned short x, y, z, pasteLoc[3];
		unsigned char block, changedblock;
		bool guestblock, ricochet = false;
		unsigned short *actionBlocks;

		memcpy(&x, buffer+1, sizeof(short));
		memcpy(&y, buffer+3, sizeof(short));
		memcpy(&z, buffer+5, sizeof(short));
		x = ntohs(x);
		y = ntohs(y);
		z = ntohs(z);

		buffer[0] = OPCODE_BLOCK_OUT;
		block = buffer[8];
		buffer[8] = buffer[7];

		if(lvl->validBlock(x, y, z) == false) {
			return;
		}

		if(buffer[7] != 0 || player->painting == true) {
			if(player->setblock == true && block == W_BLOCK_STONE)
					block = player->block;
			ricochet = true;
		}

		else block = W_BLOCK_AIR;

		changedblock = lvl->getBlock(x, y, z);
		guestblock = isGuestBlock(x, y, z);
		
		if( ((block == W_BLOCK_ROCK || changedblock == W_BLOCK_ROCK ) && player->isoperator == false) ||
			(player->userlevel < 1 && guestblock == false) ||
			(player->userlevel < 51 && isOperated == false && guestblock == false) ||
			(changedblock > 0x31 && player->userlevel < 201) ) {
				if(changedblock > 0x31)
					changedblock -= 0x31;
				buffer[7] = changedblock;
				send(player->sock, buffer, 8, 0);
				return;
		}

		if(player->densing && block != 0)
			block += 0x31;

		if(player->paste == true) {
			pasteLoc[0] = x;
			pasteLoc[1] = y;
			pasteLoc[2] = z;

			pasteBlocks(player->actionBlock_from, player->actionBlock_to, pasteLoc, true);

			sendBlock(player, player->actionBlock_from[0], player->actionBlock_from[1], player->actionBlock_from[2], W_BLOCK_LAVA, false);
			sendBlock(player, player->actionBlock_to[0], player->actionBlock_to[1], player->actionBlock_to[2], W_BLOCK_LAVA, false);

			player->paste = false;
			return;
		}

		player->lastblock = buffer[7] = block;

		if(player->action == true) {
			actionBlocks = player->actionBlock_from;

			if(player->actionBlock_which == 0 || player->actionBlock_which == 3) {
				if( player->actionBlock_which <= 1 && !(x == player->actionBlock_from[0] && 
					y == player->actionBlock_from[1] && 
					z == player->actionBlock_from[2]) )
				
					sendBlock(player, actionBlocks[0], actionBlocks[1], actionBlocks[2], 
					lvl->getBlock(actionBlocks[0], actionBlocks[1], actionBlocks[2]), false);

				if(player->actionBlock_which < 3)
					sendBlock(player, player->actionBlock_to[0], player->actionBlock_to[1], player->actionBlock_to[2], W_BLOCK_LAVA, false);

				player->actionBlock_which++;
			}

			else {
				actionBlocks = player->actionBlock_to;

				if( player->actionBlock_which <= 1 && !(player->actionBlock_to[0] == x && 
					player->actionBlock_to[1] == y && 
					player->actionBlock_to[2] == z) )

					sendBlock(player, actionBlocks[0], actionBlocks[1], actionBlocks[2], 
					lvl->getBlock(actionBlocks[0], actionBlocks[1], actionBlocks[2]), false);

				sendBlock(player, player->actionBlock_from[0], player->actionBlock_from[1], player->actionBlock_from[2], W_BLOCK_LAVA, false);

				player->actionBlock_which = 0;
			}

			actionBlocks[0] = x;
			actionBlocks[1] = y;
			actionBlocks[2] = z;
			if(buffer[8] == 0)
				sendBlock(player, actionBlocks[0], actionBlocks[1], actionBlocks[2], player->painting ? block : W_BLOCK_LAVA, false);
			return;
		}

		if(lvl->validBlock(x, y-1, z) && (block == W_BLOCK_STAIR && lvl->getBlock(x, y-1, z) == W_BLOCK_STAIR ||
			block == W_BLOCK_STAIR + 0x31 && lvl->getBlock(x, y-1, z) == W_BLOCK_STAIR + 0x31) ) {
			buffer[7] = W_BLOCK_AIR;
			send(player->sock, buffer, 8, 0);
			y--;
			y = htons(y);
			memcpy(buffer+3, &y, 2);
			y = ntohs(y);
			broadcast(buffer, 8, NULL);
			block = W_BLOCK_DOUBSTAIR;
			ricochet = true;
		}

		lvl->setBlock(x, y, z, block);

		if(buffer[7] > 0x31)
			buffer[7] -= 0x31;

		broadcast(buffer, 8, ricochet ? NULL : player);
	}

	/* Chat */
	else if(buffer[0] == OPCODE_CHAT) {
		if(player->muted == true) {
			sendChatMsg(player, false, "&eYou're muted and can not talk.");
			return;
		}

		stripChatPacket(buffer);

		for(i = 0; buffer[i] != 0; i++)
			if(buffer[i] < 0x20 || buffer[i] > 0x7E || buffer[i] == '&')
				return;
		
		if(buffer[0] == '/') {
			char paramStr[3][64], *tok;

			i = 0;
			buffer++;
			memset(paramStr, 0, sizeof(paramStr));

			Logger::add("%s commands: %s\n", player->name, buffer);

			tok = strtok(buffer, " ");

			/* Forge a paramStr out of buffer. */
			while(tok != NULL && i < 3) {
				strcpy(paramStr[i++], tok);
				tok = strtok(NULL, " ");
			}

			playerCommand(player, paramStr);
			return;
		}

		else if(buffer[0] == '@') {
			buffer++;
			privsay(player, buffer);
			return;
		}

		else if(buffer[0] == '#' && player->userlevel > 100) {
			buffer++;
			Logger::add("%s opchat: %s\n", player->name, buffer);
			opchat(player, buffer);
			return;
		}

		Logger::add("%s says: %s\n", player->name, buffer);

		if(player->hidden)
			sendChatMsgLevelSelective(false, player->userlevel, NULL, "&%c%s&f: %s", getColorCode(player->userlevel), player->name, buffer);
		else
			sendChatMsg(NULL, true, "&%c%s&f: %s", getColorCode(player->userlevel), player->name, buffer);
	}

	return;
}

void Server::processPhysicsQueue() {
	lock(mutexPhysicsQueue);
	while(physicsQueue.size() > 0) {
		physTask pTask;
		pTask = physicsQueue.front();
		physicsQueue.pop();

        if(lvl->validBlock(pTask.x, pTask.y, pTask.z) == true)
			updatePhysicsBlock(pTask);
	}
	unlock(mutexPhysicsQueue);
}

void Server::updatePhysicsBlock(physTask pTask) {
	unsigned short x, y, z;
	unsigned char type;

	x = pTask.x;
	y = pTask.y;
	z = pTask.z;
	type = pTask.block;

	if(type == W_BLOCK_WATER || type == W_BLOCK_LAVA) {
		if(lvl->validBlock(x+1, y, z) && lvl->getBlock(x+1, y, z) == W_BLOCK_AIR) { //forwards
			lvl->setBlock(x+1, y, z, type);
			sendBlock(NULL, x+1, y, z, type, true);
			pTask.x++;
			lock(mutexPhysicsQueue);
			physicsQueue.push(pTask);
			unlock(mutexPhysicsQueue);
			pTask.x--;
		}
        
		if(lvl->validBlock(x-1, y, z) && lvl->getBlock(x-1, y, z) == W_BLOCK_AIR) { //backwards
			lvl->setBlock(x-1, y, z, type);
			sendBlock(NULL, x-1, y, z, type, true);
			pTask.x--;
			lock(mutexPhysicsQueue);
			physicsQueue.push(pTask);
			unlock(mutexPhysicsQueue);
			pTask.x++;
		}
        
		if(lvl->validBlock(x, y-1, z) && lvl->getBlock(x, y-1, z) == W_BLOCK_AIR) { //down
			lvl->setBlock(x, y-1, z, type);
			sendBlock(NULL, x, y-1, z, type, true);
			pTask.y--;
			lock(mutexPhysicsQueue);
			physicsQueue.push(pTask);
			unlock(mutexPhysicsQueue);
			pTask.y++;
		}

		if(lvl->validBlock(x, y, z+1) && lvl->getBlock(x, y, z+1) == W_BLOCK_AIR) { //right
			lvl->setBlock(x, y, z+1, type);
			sendBlock(NULL, x, y, z+1, type, true);
			pTask.z++;
			lock(mutexPhysicsQueue);
			physicsQueue.push(pTask);
			unlock(mutexPhysicsQueue);
			pTask.z--;
		}

		if(lvl->validBlock(x, y, z-1) && lvl->getBlock(x, y, z-1) == W_BLOCK_AIR) { //left
			lvl->setBlock(x, y, z-1, type);
			sendBlock(NULL, x, y, z-1, type, true);
			pTask.z--;
			lock(mutexPhysicsQueue);
			physicsQueue.push(pTask);
			unlock(mutexPhysicsQueue);
			pTask.z++;
		}
	}

	else if(type == W_BLOCK_AIR) {
		if(lvl->validBlock(x+1, y, z) && (lvl->getBlock(x+1, y, z) == W_BLOCK_WATER || lvl->getBlock(x+1, y, z) == W_BLOCK_LAVA)) { //forwards
			lvl->setBlock(x+1, y, z, type);
			sendBlock(NULL, x+1, y, z, type, true);
		}
        
		if(lvl->validBlock(x-1, y, z) && (lvl->getBlock(x-1, y, z) == W_BLOCK_WATER || lvl->getBlock(x+1, y, z) == W_BLOCK_LAVA)) { //backwards
			lvl->setBlock(x-1, y, z, type);
			sendBlock(NULL, x-1, y, z, type, true);
		}
		
		if(lvl->validBlock(x, y+1, z) && (lvl->getBlock(x, y+1, z) == W_BLOCK_WATER || lvl->getBlock(x, y+1, z) == W_BLOCK_LAVA)) { //up
			lvl->setBlock(x, y+1, z, type);
			sendBlock(NULL, x, y+1, z, type, true);
		}
		
		if(lvl->validBlock(x, y-1, z) && (lvl->getBlock(x, y-1, z) == W_BLOCK_WATER || lvl->getBlock(x, y-1, z) == W_BLOCK_LAVA)) { //down
			lvl->setBlock(x, y-1, z, type);
			sendBlock(NULL, x, y-1, z, type, true);
		}
		
		if(lvl->validBlock(x, y, z+1) && (lvl->getBlock(x, y, z+1) == W_BLOCK_WATER || lvl->getBlock(x, y, z+1) == W_BLOCK_LAVA)) { //right
			lvl->setBlock(x, y, z+1, type);
			sendBlock(NULL, x, y, z+1, type, true);
		}

		if(lvl->validBlock(x, y, z-1) && (lvl->getBlock(x, y, z-1) == W_BLOCK_WATER || lvl->getBlock(x, y, z-1) == W_BLOCK_LAVA)) { //left
			lvl->setBlock(x, y, z-1, type);
			sendBlock(NULL, x, y, z-1, type, true);
		}
	}
}

void Server::sendPlayerList(Player *player) {
	std::vector<Player*>::iterator iter;
	char spawnbuf[PACKLEN_SPAWN], colorCode;

	lock(mutexPlayerList);

	for(iter = playerList.begin(); iter != playerList.end(); ++iter) {

		memset(spawnbuf, 0x20, PACKLEN_SPAWN);
		spawnbuf[0] = 0x07;

		colorCode = getColorCode((*iter)->userlevel);

		if((*iter)->id_byte == player->id_byte) {
			/* Send spawn to player who joined. */
            
			spawnbuf[1] = -1;

			spawnbuf[2] = '&';
			spawnbuf[3] = colorCode;
			memcpy(spawnbuf+4, player->name, player->namelen);

			memcpy(spawnbuf+66, &player->posx, 2);
			memcpy(spawnbuf+68, &player->posy, 2);
			memcpy(spawnbuf+70, &player->posz, 2);
			spawnbuf[72] = player->rotx;
			spawnbuf[73] = player->roty;

			send(player->sock, spawnbuf, PACKLEN_SPAWN, 0);

			spawnbuf[1] = player->id_byte;
			broadcast(spawnbuf, PACKLEN_SPAWN, player);

			sendChatMsg(NULL, true, "&e%s joined the game!", player->name);
		}

		else {
			/* Send player at playerList[i] to player. */
		    if((*iter)->hidden && ((*iter)->userlevel > player->userlevel))
			     continue; /* Don't show hidden players */
			spawnbuf[1] = (*iter)->id_byte;
			spawnbuf[2] = '&';
			spawnbuf[3] = colorCode;
			memcpy(spawnbuf+4, (*iter)->name, (*iter)->namelen);
			memcpy(spawnbuf+66, &(*iter)->posx, 2);
			memcpy(spawnbuf+68, &(*iter)->posy, 2);
			memcpy(spawnbuf+70, &(*iter)->posz, 2);
			spawnbuf[72] = (*iter)->rotx;
			spawnbuf[73] = (*iter)->roty;
			send(player->sock, spawnbuf, PACKLEN_SPAWN, 0);
		}
	}

	unlock(mutexPlayerList);
}

void Server::sendMap(Player *player) {
	unsigned long sent, left, total;
	unsigned short chunksize;
	unsigned char mapbuf[PACKLEN_MAP_CHUNK];
	unsigned char percentage;

	lock(mutexSendMap);

	if(currentlySending < 1) {
		mapSend = (unsigned char*) malloc(lvl->getCompressedMapLen() + 16384);
		lvl->lvlLock();
		lvl->setCompressedMapLen(gzip_compress(mapSend, lvl->getServerLevel_safe(), lvl->getDecompressedMapLen() - OUR_HEADERLEN));
		lvl->freeServerLevel_safe();
		lvl->lvlUnlock();
	}

	currentlySending++;
	unlock(mutexSendMap);

	/* 0x02 = OPCODE_MAP_BEGIN */
	send(player->sock, "\x02", PACKLEN_MAP_BEGIN, 0);

	sent = 0;
	total = lvl->getCompressedMapLen();

	mapbuf[0] = OPCODE_MAP_CHUNK;

	while(sent < total) {
		left = total - sent;
		chunksize = (left < 1024 ? (unsigned short) left : 1024);

		memcpy(mapbuf+3, mapSend+sent, chunksize);
		sent += chunksize;

		percentage = (unsigned char) ((sent * 100) / total);
		mapbuf[3+chunksize] = percentage;

		/* Weird doing this last but it reduce one call to htons. */
		chunksize = htons(chunksize);
		memcpy(mapbuf+1, &chunksize, sizeof(short));
		send(player->sock, (const char*) mapbuf, PACKLEN_MAP_CHUNK, 0);

		msleep(10);
	}

	mapbuf[0] = OPCODE_MAP_FINISH;

	chunksize = lvl->getDimX();
	memcpy(mapbuf+1, &chunksize, sizeof(short));

	chunksize = lvl->getDimY();
	memcpy(mapbuf+3, &chunksize, sizeof(short));

	chunksize = lvl->getDimZ();
	memcpy(mapbuf+5, &chunksize, sizeof(short));

	send(player->sock, (const char*) mapbuf, PACKLEN_MAP_FINISH, 0);

	lock(mutexSendMap);
	currentlySending--;
	if(currentlySending < 1) {
		free(mapSend);
		mapSend = NULL;
	}
	unlock(mutexSendMap);
}

void Server::sendServerInfo(Player *player) {
	static char infoBuffer[PACKLEN_LOGIN];
	static bool doneThis = false;

	/* Space out these strings ONCE only. */
	if(doneThis == false) {
		memset(infoBuffer, 0x20, sizeof(infoBuffer));

		infoBuffer[0] = 0x00;
		infoBuffer[1] = 0x07;

		memcpy(infoBuffer+2, srvName, strlen(srvName));
		memcpy(infoBuffer+2+64, srvMotd, strlen(srvMotd));

		doneThis = true;
	}

	/* > 50 = > TRUSTEDS ONLY */
	infoBuffer[2+64+64] = (player->userlevel > 100 ? 0x64 : 0x00);

	send(player->sock, infoBuffer, PACKLEN_LOGIN, 0);
}

bool Server::isBanned(char *name, char *ip) {
	std::vector<Banned>::iterator iter;

	for(iter = bannedList.begin(); iter != bannedList.end(); iter++) {
		if(strcmp(iter->name, name) == 0 || strcmp(iter->ip, ip) == 0)
			return true;
	}

	return false;
}

Player *Server::getPlayerByName(char *playerName) {
	std::vector<Player*>::iterator iter;
	char playerCompare1[32], playerCompare2[32];
	Player *ret = NULL;

	if(strlen(playerName) > 32)
		return 0;

	strcpy(playerCompare1, playerName);
	toLowerCase(playerCompare1);
	
	lock(mutexPlayerList);

	for(iter = playerList.begin(); iter != playerList.end(); ++iter) {
		strcpy(playerCompare2, (*iter)->name);
		toLowerCase(playerCompare2);

		if(strcmp(playerCompare2, playerCompare1) == 0) {
			ret = (*iter);
			break;
		}
	}

	unlock(mutexPlayerList);
	return ret;
}

Player *Server::getPlayerById(unsigned char id_byte) {
	std::vector<Player*>::iterator iter;
	Player *ret = NULL;

	lock(mutexPlayerList);

	for(iter = playerList.begin(); iter != playerList.end(); ++iter) {
		if((*iter)->id_byte == id_byte) {
			ret = (*iter);
			break;
		}
	}

	unlock(mutexPlayerList);
	return ret;
}
