#ifndef _SERVER_H
#define _SERVER_H

class Level;

struct Task;
struct physTask;
struct User;
struct Banned;
struct Player;

class Server {

public:

	Server();
	~Server();

	bool step();

private:

	static void *thrExecTask(void *arg) {
		Server *_this;
		_this = (Server*) arg;
		while(_this->execTask());
		LeaveThread(0);
	}

	bool execTask();

	void handleLogin(Task *task);
	void sendPlayerList(Player *player);
	void sendMap(Player *player);
	void sendServerInfo(Player *player);
	bool verifyName(char *name, char *hash);

	void parsePacket(Task *task);
	bool playerCommand(Player *player, char paramStr[3][64]);
	bool privsay(Player *player, char *buffer);
	bool opchat(Player *player, char *buffer);

	void beat();
	void broadcast(char *packet, int paklen, Player *exclude);
	void broadcastLevelSelective(char *packet, int paklen, bool toLowers, unsigned char level, Player *exclude);

	void fakeLog(Player *player,  bool out);
	void teleportPlayer(Player *toPlayer, Player *who);

	void kickPlayer(Player *player, const char *kickMsg, bool announce);
	void banPlayer(char *playerName, const char *banMsg, bool unban);
	void releasePlayer(Player *player);

	void refreshColor(Player *player);

	void drawBlocks(unsigned short *drawFrom, unsigned short *drawTo, unsigned char block, bool overwriteDense, Player *player);
	bool changeBlockState(bool toDense, unsigned short *drawFrom, unsigned short *drawTo);
	bool pasteBlocks(unsigned short *drawFrom, unsigned short *drawTo, unsigned short *pasteLoc, bool samemap);
	
	void sendBlock(Player *player, unsigned short x, unsigned short y, unsigned short z, unsigned char block, bool broad);
	
	void stripChatPacket(char *packet);
	void sendChatMsg(Player *player, bool broad, const char *format, ...);
	void sendChatMsgLevelSelective(bool toLowers, unsigned char level, Player *exclude, const char *format, ...);

	bool isBanned(char *name, char *ip);
	bool setUser(char *playerName, unsigned char level, Player *changer);
	bool isOperatorOn();
	void setGuestArea();
	bool isGuestBlock(unsigned short x, unsigned short y, unsigned short z);

	void processPhysicsQueue();
	void updatePhysicsBlock(physTask physicsTask);

	User *getUserByName(char *playerName);
	Player *getPlayerByName(char *playerName);
	Player *getPlayerById(unsigned char id_byte);
	const char *levelToClass(unsigned char userlevel);
	char getColorCode(unsigned char userlevel);

	int playerNum;
	int playerMax;
	int backupInterval;
	int salt;

	bool srvGuest;
	unsigned short guestFrom[3];
	unsigned short guestTo[3];

	char srvName[64];
	char srvMotd[64];
	char srvPort[8];

	bool srvPublic;
	bool srvVerify;
	bool isOperated;
	
	time_t timer;
	time_t time_beat;
	time_t time_savemap;
	time_t time_backupmap;

	int sock_listener;
	int nfds;
	struct timeval timeout;
	struct addrinfo *listener;
	struct addrinfo *heartbeat;
	fd_set master, readfds;
#ifdef WIN32
	WSADATA wsaData;
#endif

	mutex_t mutexPlayerList;
	std::vector<Player*> playerList;

	mutex_t mutexThreadData;
	int threadsNeeded;
	int threadsRunning;
	thread_t *threads;
	std::queue<Task*> taskQueue;

	mutex_t mutexPhysicsQueue;
	std::queue<physTask> physicsQueue;

	std::vector<User> userList;
	std::vector<Banned> bannedList;

	std::stack<unsigned char> idList;

	Level *lvl;
	Level *lvl_restore;
	unsigned char *mapSend;
	char currentlySending;
	mutex_t mutexSendMap;

};

#endif
