#ifndef _DEFINES_H
#define _DEFINES_H

#define PROTO_VERSION 0x07
#define OUR_HEADERLEN 17

#define OPCODE_BLOCK_IN 0x05
#define OPCODE_BLOCK_OUT 0x06
#define PACKLEN_BLOCK_IN 9
#define PACKLEN_BLOCK_OUT 8

#define OPCODE_MOVE 0x08
#define PACKLEN_MOVE 10

/* These will be used for optimizations later on! */
#define OPCODE_MOVCHANGE_OUT 0x0A
#define OPCODE_ROTCHANGE_OUT 0x0B

#define OPCODE_CHAT 0x0D
#define PACKLEN_CHAT 66

#define OPCODE_LOGIN 0x00
#define OPCODE_VERSION 0x07
#define PACKLEN_LOGIN 131

#define OPCODE_KICK 0x0E
#define PACKLEN_KICK 65

#define OPCODE_SPAWN 0x07
#define PACKLEN_SPAWN 74

#define OPCODE_MAP_BEGIN 0x02
#define OPCODE_MAP_CHUNK 0x03
#define OPCODE_MAP_FINISH 0x04
#define PACKLEN_MAP_BEGIN 1
#define PACKLEN_MAP_CHUNK 1028
#define PACKLEN_MAP_FINISH 7

#define WIN32

#ifdef WIN32

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <winsock2.h>
#include <ws2tcpip.h>

#define LaunchThread(THRID, PROC, PARAM) CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) PROC, (LPVOID) PARAM, 0, THRID)
#define LeaveThread(RET) ExitThread(RET)
#define newmutex(MUTEX) MUTEX = CreateMutex(NULL, false, NULL)
#define lock(MUTEX) WaitForSingleObject(MUTEX, INFINITE)
#define trylock(MUTEX) WaitForSingleObject(MUTEX, 0)
#define unlock(MUTEX) ReleaseMutex(MUTEX)
#define closesock(sock) closesocket(sock)
#define msleep(msec) Sleep(msec)

#define MUTEX_BUSY WAIT_TIMEOUT
typedef DWORD thread_t;
typedef HANDLE mutex_t;
typedef int socklen_t;

#else

#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <netdb.h>
#include <unistd.h>
#include <pthread.h>

#define LaunchThread(THRID, PROC, PARAM) pthread_create(THRID, NULL, PROC, (void*) PARAM)
#define LeaveThread(RET) pthread_exit(NULL)
#define newmutex(MUTEX) pthread_mutex_init(&MUTEX, NULL)
#define lock(MUTEX) pthread_mutex_lock(&MUTEX)
#define trylock(MUTEX) pthread_mutex_trylock(&MUTEX)
#define unlock(MUTEX) pthread_mutex_unlock(&MUTEX)
#define closesock(sock) close(sock)
#define msleep(msec) usleep(msec*1000)

#define MUTEX_BUSY EBUSY
typedef pthread_t thread_t;
typedef pthread_mutex_t mutex_t;

#endif

#endif
